/* ihx.c */
/* Intel Hex Format load library */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ihx.h"
#include "pic_db.h"

static char rcsid[] = "$Id: ihx.c,v 1.3 1997/12/28 15:23:13 kondou Exp kondou $";

extern int have_prog;
extern int have_eedata;
extern int have_fuse;

static int ihx8m_line(unsigned char * line, unsigned short * data,
					  int * len, int * loc);

#define BUF_SIZE 1024

/* read ihx8m file to pic memory buffer */
int pic_ihx8m_load(FILE * fp, int pic_type,
				   unsigned short * fuse,
				   unsigned short * id,
				   unsigned short * prog,
				   unsigned short * data)
{
  unsigned char buf[BUF_SIZE];
  unsigned short buf_tmp[BUF_SIZE];
  int len, loc;
  int line_count = 0;

  while(fgets(buf, BUF_SIZE, fp)){
	if (!strcmp(buf, ":00000001FF\n") || !strcmp(buf, ":00000001FF\r\n"))
	  {  /* end of data */
	  break;
	}
	if (ihx8m_line(buf, buf_tmp, &len, &loc) != 0) {
	  fprintf(stderr, " ihx8m_load: format or check sum error(%d): %s\n",
			  line_count+1, buf);
	  return -1;
	}
	if (loc >= DATA_FROM){  /* data*/
	  have_eedata = 1;
	  memcpy(data+loc-DATA_FROM, buf_tmp, len*sizeof(unsigned short));
	} else if (loc == PIC_DB[pic_type].fuse_loc){ /* fuse */
	  have_fuse = 1;
	  *fuse = buf_tmp[0];
	} else if (PIC_DB[pic_type].id_loc == loc){
	  memcpy(id, buf_tmp, 3);
	} else {   /* program */
	  have_prog = 1;
	  memcpy(prog+loc, buf_tmp, len*sizeof(unsigned short));
	}
	line_count++;
  }
#if DEBUG
  printf("\nPROGRAM");
  for(len = 0; len < PIC_DB[pic_type].prog_size; len++){
	if ((len % 8) == 0) printf("\n%04x: ", len);
	printf("%04x ", prog[len]);
  }
  printf("\nfuse = %02x", *fuse);
  printf("\nEEPROM");
  for(len = 0; len < PIC_DB[pic_type].data_size; len++){
	if ((len % 16) == 0) printf("\n%04x: ", len);
	printf("%02x(%c) ", data[len],(char)data[len]);
  }
#endif /* DEBUG */

  return 0;
}


/* IHX8M $B$N0l9T$rFI$_9~$`(B */
static int ihx8m_line(unsigned char * line,
			   unsigned short * data,
			   int * len, int * loc)
{

  unsigned char * cp;
  unsigned char buf[BUF_SIZE];
  int d1, d2;
  int i;
  int check = 0;

  cp = line;
  if ( *cp != ':' ){
	fprintf(stderr, "format error: %s(%d) => %s", __FILE__, __LINE__, cp);
	return -1;
  } /* not format */
#if DEBUG
  printf("\nline=%s", line);
#endif

  cp ++;
  strncpy(buf, cp, 2);
  buf[2] = 0;
  if ( sscanf(buf, "%02X", len) != 1 ) {
	fprintf(stderr, "format error: %s(%d) => %s", __FILE__, __LINE__, cp);
	return -1;
  }
  *len /= 2;
#if DEBUG
  printf("len=%x, %s\n", *len, buf);
#endif

  cp += 2;           /* address of this line */
  strncpy(buf, cp, 4);
  buf[4] = 0;
  if ( sscanf(buf, "%04X", loc) != 1 ) {
	fprintf(stderr, "format error: %s(%d) => %s", __FILE__, __LINE__, cp);
	return -1;
  }
  *loc /= 2;
#if DEBUG
  printf("loc=%x, %s\n", *loc, buf);
#endif

  check += ((2* *loc) & 0xff) + (((2* *loc) >> 8) & 0xff) + 2* *len;

  cp += 4;           /* 00 fixed */
  strncpy(buf, cp, 2);
  buf[2] = 0;
  if ( strcmp(buf, "00") ) {
	fprintf(stderr, "format error: %s(%d) => %s", __FILE__, __LINE__, cp);
	return -1;
  }

  cp += 2;           /* data of this line */
  for(i = 0; i < *len ; i++){
	strncpy(buf, cp, 2);
	buf[2] = 0;
	if ( sscanf(buf, "%02X", &d1) != 1) {
	  fprintf(stderr, "format error: %s(%d) => %s, %s\n",
			  __FILE__, __LINE__, cp, buf);
	  return -1;
	}
	cp += 2;
	strncpy(buf, cp, 2);
	buf[2] = 0;
	if ( sscanf(buf, "%02X", &d2) != 1) {
	  fprintf(stderr, "format error: %s(%d) => %s, %s\n",
			  __FILE__, __LINE__, cp, buf);
	  return -1;
	}
	cp += 2;
	data[i] = (unsigned short)((d1 & 0x00ff) + ((d2 & 0x00ff)<< 8));
	check += (d1 & 0x00ff) + (d2 & 0x00ff);
#if DEBUG
  printf("data=%04x\n", data[i]);
#endif
  }

  strncpy(buf, cp, 2);   /* check sum */
  buf[2] = 0;
  if ( sscanf(buf, "%02X", &d1) != 1 ) {
	fprintf(stderr, "format error: %s(%d) => %s, %s\n",
			__FILE__, __LINE__, cp, buf);
	return -1;
  }
  if (d1 != ((-check) & 0x00ff)) {
	fprintf(stderr, "check sum error: %s(%d) => %s, d1=%x, check = %x(%x)\n",
			__FILE__, __LINE__, cp, d1, check, (-check)&0x00ff);
	return -2;
  } /* check sum error */
#if DEBUG
  printf("check=%x, %s\n", d1, buf);
#endif

  cp += 2;
  if (*cp != '\n' && *cp != '\r') {
	fprintf(stderr, "format error: %s(%d) => %s\n", __FILE__, __LINE__, cp);
	return -1;
  } /* line shoud be LF */
  return 0;
}

