/* io_lp.h */
/* pararell port access utility */

#ifndef IO_LP_H
#define IO_LP_H

#ifndef __FreeBSD__
#define outbyte(data, io) outb((data), (io))
#include <asm/io.h>
#else
#define outbyte(data, io) outb((io), (data))
#include <machine/cpufunc.h>
#endif

#include <unistd.h>
#include <errno.h>

/* I/O port address */
#define LP_DATA_REG    0x378
#define LP_STATUS_REG  0x379
#define LP_CONTROL_REG 0x37a

/* status regs */
#define LP_BUSY        (1 << 7)
#define LP_N_ACK       (1 << 6)
#define LP_PE          (1 << 5)
#define LP_SELECT      (1 << 4)
#define LP_ERROR       (1 << 3)

/* control regs */
#define LP_IO          (1 << 5)
#define LP_INT         (1 << 4)
#define LP_N_SLCTIN    (1 << 3)
#define LP_N_INIT      (1 << 2)
#define LP_AUTOFD      (1 << 1)
#define LP_N_STROBE    (1)

/* LP access macros */
#ifndef __FreeBSD__
#define LP_INIT(turn_on) \
do{ \
	  if (ioperm(LP_DATA_REG,1,turn_on)){perror("ioperm");exit(1);} \
	  if (ioperm(LP_STATUS_REG,1,turn_on)){perror("ioperm");exit(1);} \
	  if (ioperm(LP_CONTROL_REG,1,turn_on)){perror("ioperm");exit(1);} \
  }while(0)
#else
#define LP_INIT(turn_on)
#endif

#define LP_WRITE_CONTROL(byte) outbyte(byte, LP_CONTROL_REG)
#define LP_READ_STATUS() inb(LP_STATUS_REG)
#if DEBUG_DRY
#define LP_WRITE_DATA(byte) outbyte(byte, LP_DATA_REG),fprintf(stdout, "%d%d%d%d%d%d%d%d\n", (((byte)&0x80)!=0),(((byte)&0x40)!=0),(((byte)&0x20)!=0),(((byte)&0x10)!=0),(((byte)&0x08)!=0),(((byte)&0x04)!=0),(((byte)&0x02)!=0),(((byte)&0x01)!=0))
#else
#define LP_WRITE_DATA(byte) outbyte(byte, LP_DATA_REG)
#endif
#define LP_READ_DATA() inb(LP_DATA_REG)

#endif /* IO_LP_H */
