/* picwriter */
/* Writer of PIC Micro controler.
   support devices are
   PIC16[C|F]84
   PIC12C509

   Copyright (C) 1997, KONDOU Nobuhiro <nkon@osk2.3web.ne.jp>
*/

static char rcsid[] = "$Id: picwriter.c,v 1.4 1997/12/31 04:51:46 kondou Exp kondou $";

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#define MAIN 1

#include "ihx.h"
#include "writer.h"
#include "io_lp.h"
#include "pic_db.h"

#define OPTION_FORMAT "crv::ht:f:"
#define USAGE_OUTPUT  stderr
#define DEFAULT_VERBOSE 1

int Verbose = DEFAULT_VERBOSE;
int PORT = 0;
int opt_c = 0;
int opt_r = 0;
int have_prog = 0;
int have_eedata = 0;
int have_fuse = 0;
char * ProgramName = 0;

#ifdef __FreeBSD__
static FILE *iofp = (FILE*)NULL;
#endif

void usage(void)
{
  fprintf(USAGE_OUTPUT, "picwriter Ver. 1.0      Last-modified: 1997.12/28\n");
  fprintf(USAGE_OUTPUT, "Copyright (C) "
		  "KONDOU Nobuhiro <nkon@osk2.3web.ne.jp>\n");
  fprintf(USAGE_OUTPUT, "Usage: picwriter -[options] filename\n");
  fprintf(USAGE_OUTPUT, "Options:\n");
  fprintf(USAGE_OUTPUT, "\t-v[0|1|2]       : verbose level. "
		"[default %d]\n",  DEFAULT_VERBOSE);
  fprintf(USAGE_OUTPUT, "\t-t target_type  : pic16c84, pic16f84, pic12c509.  "
		  "[defualt=pic16c84]\n");
  fprintf(USAGE_OUTPUT, "\t-f source_format: ihx8m, ihx16, ihxlh, ihxhl. "
		  "[default=ihx8m]\n");
  fprintf(USAGE_OUTPUT, "\t-c              : chip erase and exit.\n");
  fprintf(USAGE_OUTPUT, "\t-r              : dump chip and exit.\n");
  fprintf(USAGE_OUTPUT, "\t-h              : print this message.\n");
  exit(1);
}

void myexit(void)
{
  prog_verify_mode(&PORT, OFF);
  LP_WRITE_DATA(0);
  LP_INIT(0);  /* disable pararell port */
#ifdef __FreeBSD__
  if(iofp != (FILE*)NULL)
    fclose(iofp);
#endif

}  

void mysetup(void)  /* setup myself */
{
  atexit(myexit);
  LP_INIT(1);  /* enable pararell port */
  pic_wait_setup();
}


/* ץȥǡѤΥХåեݤ롣
    pid_db.h 롣*/
void malloc_mems(int pic_type, unsigned short ** prog, unsigned short ** data)
{
  if ((*prog = (unsigned short *)
	   (malloc(sizeof(unsigned short) * PIC_DB[pic_type].prog_size))) == NULL){
	fprintf(stderr,
			"%s: Not enouch memory(%s:%d)", ProgramName, __FILE__, __LINE__);
	exit(1);
  }
  memset(*prog, 0xff, PIC_DB[pic_type].prog_size * sizeof(unsigned short));
  if ((*data = (unsigned short *)
	   (malloc(sizeof(unsigned short) * PIC_DB[pic_type].data_size))) == NULL){
	fprintf(stderr,
			"%s: Not enouch memory(%s:%d)", ProgramName, __FILE__, __LINE__);
	exit(1);
  }
  memset(*data, 0xff, PIC_DB[pic_type].data_size * sizeof(unsigned short));
}

/* main program */
void main(int argc, char ** argv)
{
  int format = IHX8M;
  int pic_type = PIC16C84;
  FILE * fp = 0;
  int c;
  unsigned short fuse;
  unsigned short id[3];
  unsigned short * prog=0;
  unsigned short * data=0;
  int (*load_fp)(FILE*, int , unsigned short *, unsigned short *,
				 unsigned short *, unsigned short *) = pic_ihx8m_load;

  if (argc < 2){
	usage();
	exit(1);
  }

  ProgramName = strdup(argv[0]);

  /* parse option */
  while((c = getopt(argc, argv, OPTION_FORMAT)) != EOF){
	switch(c){
	case 'c': /* chip erase */
	  if (Verbose > 1){
		printf("Do chip erase.\n");
	  }
	  opt_c = 1;
	  break;
	case 'r': /* dump chip */
	  if (Verbose > 1){
		printf("dump chip only.\n");
	  }
	  opt_r = 1;
	  break;
	case 't': /* pic type */
	  if (!strcasecmp("pic16c84", optarg)){
		pic_type = PIC16C84;
		if (Verbose > 1){
		  printf("pic_type = PIC16C84\n");
		}
	  } else if (!strcasecmp("pic16f84", optarg)){
		pic_type = PIC16F84;
		if (Verbose > 1){
		  printf("pic_type = PIC16F84\n");
		}
	  } else if (!strcasecmp("pic12c509", optarg)){
		pic_type = PIC12C509;
		if (Verbose > 1) {
		  printf("pic_type = PIC12C509\n");
		}
	  } else {
		fprintf(stderr, "%s: option -t has these parameter\n"
				"\tpic16c84, pic16f84, pic12c509\n", ProgramName);
		exit(1);
	  }
	  break;
	case 'f': /* file format */
	  if (!strcasecmp("ihx8m", optarg)){
		format = IHX8M;
		load_fp = pic_ihx8m_load;
		if (Verbose > 1){
		  printf("format = IHX8M\n");
		}
	  } else if (!strcasecmp("ihx16", optarg)){
		format = IHX16;
		load_fp = pic_ihx8m_load;
		if (Verbose > 1) {
		  printf("format = IHX16\n");
		}
	  } else {
		fprintf(stderr, "%s: option -f has these parameter\n"
				"\tihx8m, ihx16\n", ProgramName);
		exit(1);
	  }
	  break;
	case 'v': /* verbose level */
	  if (!optarg){
		Verbose = 2;
	  } else if(!strcmp(optarg, "0")){
		Verbose = 0;
	  } else if(!strcmp(optarg, "1")){
		Verbose = 1;
	  } else if(!strcmp(optarg, "2")){
		Verbose = 2;
	  } else {
		Verbose = 2;
	  }
	  break;
	case 'h': /* help */
	  usage();
	  break;
	case ':': /* missing paramater */
	  /* getopt has defualt error message */
	  break;
	case '?': /* unknown option char */
	  /* getopt has defualt error message */
	  break;
	default:
	  fprintf(stderr, "BUG: %s(%d)\n", __FILE__, __LINE__);
	  exit(1);
	}
  }

#ifdef __FreeBSD__
  if((iofp = fopen("/dev/io", "r")) == NULL) {
    fprintf(stderr, "You donot have permition to access I/O.\n");
    exit(1);
  }
#endif

  if (opt_r){ /* dump chip and exit */
	int len;

	mysetup();
	malloc_mems(pic_type, &prog, &data);

	/* read program */
	pic_read(&PORT, PROGRAM_MEMORY, prog, PIC_DB[pic_type].prog_size,pic_type);
	/* read memory */
	pic_read(&PORT, DATA_MEMORY, data, PIC_DB[pic_type].data_size, pic_type);
	printf("\nPROGRAM");
	for(len = 0; len < PIC_DB[pic_type].prog_size; len++){
	  if ((len % 8) == 0) printf("\n%04x: ", len);
	  printf("%04x ", prog[len]);
	}
	printf("\nEEPROM");
	for(len = 0; len < PIC_DB[pic_type].data_size; len++){
	  if ((len % 16) == 0) printf("\n%04x: ", len);
	  printf("%02x(%c) ", data[len],(char)data[len]);
	}
	exit(0);
  }

  if (opt_c) { /* chip erase only */
	mysetup();
	pic_chip_erase(&PORT, pic_type);
	exit(0);
  }

  /* open data file */
  if (optind == (argc - 1)) {
	if (!strcmp(argv[optind],"-")){
	  fp = stdin;
	} else if ((fp = fopen(argv[optind], "r")) == NULL){
	  perror(argv[optind]);
	  exit(1);
	}
  } else {
	fprintf(stderr, "%s: only one data file should be passed.\n", ProgramName);
	usage();
	exit(1);
  }

  malloc_mems(pic_type, &prog, &data);

  /* load data file */
  if (load_fp(fp, pic_type, &fuse, id, prog, data) < 0){
	fprintf(stderr, "file format error: %s\n", argv[optind]);
	exit(1);
  }
  fclose(fp);
	
  mysetup();

  pic_chip_erase(&PORT, pic_type); /* chip erase at first */


  if(have_prog)   /* write program */
    pic_write(&PORT, PROGRAM_MEMORY,
	      prog, PIC_DB[pic_type].prog_size, pic_type);
  if(have_eedata) /* write memory */
    pic_write(&PORT, DATA_MEMORY, data, PIC_DB[pic_type].data_size, pic_type);
  if(have_fuse)   /* write fuse */
    pic_write_fuse(&PORT, fuse, pic_type);
	
  free(prog);
  free(data);

  exit(0);
}

