/* writer.c */
/* writer routine for PIC micro controler */

#define DEBUG_DRY 1
#undef DEBUG_DRY

#include <stdio.h>
#include "io_lp.h"
#include "writer.h"
#include "pic_db.h"

#define CYCLE_BIT 14

extern int Verbose;

/* delay functions */

static int _pic_wait_setup=0;

void pic_wait_setup(void)
{
#ifndef __FreeBSD__
  if (_pic_wait_setup){	return;  }
  /* 0x80 ˽Ϥ */
  ioperm(0x80,1,1);
#endif
  _pic_wait_setup=1;
}

/* microseconds Ԥ */
extern __inline__ void pic_wait(int microseconds)
{
  int i;
  for(i=0; i < microseconds; i++){
	outbyte(0x80, 0x80);
  }
}

/* p_or_mPROGRAM_MEMORYλdata14bitڤͤ(pic_type==pic16c84)
   p_or_mDATA_MEMORYλdata8bitڤͤ */
static unsigned short trunc_data(unsigned short data, int p_or_m, int pic_type)
{
  if (p_or_m == DATA_MEMORY){
	/* data memory is always 8bit */
	data &= (0xffff >>(sizeof(data)*8-PIC_DB[pic_type].sizeof_dataword));
  } else {
	/* program memory is 14 or 12bit */
	data &= (0xffff >>(sizeof(data)*8-PIC_DB[pic_type].sizeof_progword));
  }
  return data;
}

/* bit ΩäƤ data ѹ */
int lp_pic_modify(int * port, int bit, int data)
{
  *port = (*port & (~bit)) + bit*(data != 0);
  return *port;
}

/* data==1 λåΩ夲롣data==0 λΩ */
/* clock enable  L(ENABLE) Τޤ */
void lp_pic_clock(int *port, int data)
{
  lp_pic_modify(port, LP_PIC_CLK, (data != 0));
  LP_WRITE_DATA(((unsigned char)(*port)));
/*  lp_pic_modify(port, LP_PIC_CLK_ENABLE, (data == 0));
  LP_WRITE_DATA(((unsigned char)(*port)));
*/
  lp_pic_modify(port, LP_PIC_CLK_ENABLE, DOWN);
  LP_WRITE_DATA(((unsigned char)(*port)));
  pic_wait(1);
}

/* data==1 λǡΩ夲롣data==0 λΩ */
/* data enable  L(ENABLE) Τޤ */
void lp_pic_data_out(int *port, int data)
{
  lp_pic_modify(port, LP_PIC_DATA_ENABLE, 0);
  LP_WRITE_DATA(((unsigned char)(*port)));
  lp_pic_modify(port, LP_PIC_DATA, (data != 0));
  LP_WRITE_DATA(((unsigned char)(*port)));
/*  lp_pic_modify(port, LP_PIC_DATA_ENABLE, (data == 0));
  LP_WRITE_DATA(((unsigned char)(*port)));
*/
  lp_pic_modify(port, LP_PIC_DATA_ENABLE, DOWN);
  LP_WRITE_DATA(((unsigned char)(*port)));
  pic_wait(1);
}  

/* ǡɤ߹ 1 or 0 ֤ */
/* data enable  H(DISABLE) Τޤ */
int lp_pic_data_in(int *port)
{
  lp_pic_modify(port, LP_PIC_DATA_ENABLE, UP);
  LP_WRITE_DATA(((unsigned char)(*port)));
#ifndef USE_BUSY
  return ((LP_READ_STATUS() & LP_PIC_DATA_READ) != 0);
#else
  return (((~LP_READ_STATUS()) & LP_PIC_DATA_READ) != 0);
#endif
}  

/* set MCLR voltage */
/* cmd  write.h Ƥ롢PIC_MCLR_VDD/VPP/ZERO */
void lp_pic_mclr(int *port, int cmd)
{
  lp_pic_modify(port, LP_PIC_MCLR_1,((cmd & LP_PIC_MCLR_1) != 0));
  lp_pic_modify(port, LP_PIC_MCLR_2,((cmd & LP_PIC_MCLR_2) != 0));
  LP_WRITE_DATA(((unsigned char)(*port)));
}

/* send 6 cycles command sign to PIC (data is LSB first) */
/* data  PIC_CMD_* */
void send_sign(int *port, int data)
{
  int i;

#if DEBUG_DRY
  printf("SendSign(%02x)\n", data);
#endif

  for(i = 0; i < 6; i++){
	lp_pic_clock(port, UP);
	lp_pic_data_out(port, ((data & 0x01)!=0));
	lp_pic_clock(port, DOWN);
	data >>= 1;
  }
  pic_wait(1);
}

/* load data to PIC */
/* data is already trunc_data() */
/* send_sign(PIC_CMD_LOAD_PROG|PIC_CMD_LOAD_DATA) ˤĤŤơ
   ǡ롣ǡ14ӥå뤬 0  */
void load_cycle(int * port, int data, int cmd, int pic_type)
{
  int i;

#if DEBUG_DRY
  printf("Load Cycle\n");
#endif

  send_sign(port, cmd);

  /* ꡼ǥ󥰥ӥåȤ */
  lp_pic_data_out(port, DOWN);
  lp_pic_clock(port, UP);
  lp_pic_clock(port, DOWN);
  
  for(i = 0; i < CYCLE_BIT; i++){
	lp_pic_clock(port, UP);
	lp_pic_data_out(port, ((data & 0x01) != 0));
	lp_pic_clock(port, DOWN);
	data >>=1;
  }
  /* λӥåȤ */
  lp_pic_clock(port, UP);
  lp_pic_data_out(port, DOWN);
  lp_pic_clock(port, DOWN);

  pic_wait(1);

}

/* read data from PIC */
/* data should be trunc_data() */
/* send_sign(PIC_READ_PROG|PIC_READ_DATA) ˤĤŤƥǡɤߤȤ */
int read_cycle(int * port, int cmd, int pic_type)
{
  int i;
  int data = 0;

#if DEBUG_DRY
  printf("Read Cycle\n");
#endif
  send_sign(port, cmd);     /* read signal */

  lp_pic_clock(port, UP);
  lp_pic_clock(port, DOWN);
  
  for(i = 0; i < CYCLE_BIT; i++){
	lp_pic_clock(port, UP);
	lp_pic_clock(port, DOWN);
	data |= (lp_pic_data_in(port) & 0x01) << i;
  }
  lp_pic_clock(port, UP);
  lp_pic_clock(port, DOWN);

  pic_wait(1);

  return data;

}

/***** high level interface ********/

void prog_verify_mode(int * port, int on)
{
#if DEBUG_DRY
	printf("prog_verify_mode(%d)", on);
#endif

  lp_pic_mclr(port, PIC_MCLR_ZERO);
  lp_pic_data_out(port, DOWN);
  lp_pic_clock(port, DOWN);
  pic_wait(1000);

  if (on){
	lp_pic_mclr(port, PIC_MCLR_VPP);
	if (Verbose > 1){
	  printf("Program/Verify mode ON\n");
	}
  } else {
	lp_pic_mclr(port, PIC_MCLR_ZERO);
	if (Verbose > 1){
	  printf("Program/Verify mode OFF\n");
	}
  }
}

/* write program or memory data */
void pic_write(int * port, 
			   int p_or_m,   /* program memory or data memory */
			   unsigned short * data,  /* data buffer */
			   int len,      /* data length */
			   int pic_type) /* pic device */
{
  int i;
  int load_cmd = PIC_CMD_LOAD_PROG;
  int read_cmd = PIC_CMD_READ_PROG;
  unsigned short readed;
  int cur_addr=0;

  if (Verbose > 0){
	printf("Writing %s Memory...\n", (p_or_m ? "EEPROM" : "PROG"));
  }

  prog_verify_mode(port, ON); /* MCLR=Vpp program/verify mode */
  
  /* if p_or_m==1 then data memory else program memory*/
  if (p_or_m == DATA_MEMORY){
	load_cmd = PIC_CMD_LOAD_DATA;
	read_cmd = PIC_CMD_READ_DATA;
  }

  for(i = 0; i < len; i++){
	load_cycle(port, trunc_data(*data, p_or_m, pic_type), load_cmd, pic_type);
	send_sign(port, PIC_CMD_BEGIN_PROG);
	pic_wait(10*1000);
	if (Verbose > 1){
	  printf("Address: %04x\n", cur_addr);
	  printf("Write %s: %04x\n", (p_or_m ? "EEPROM" : "PROG"), *data);
	} else if (Verbose > 0){
	  printf("Address: %04x %s: %04x\r",
			 cur_addr, (p_or_m ? "EEPROM" : "PROG"), *data);
	}
	/* verify */
	readed = trunc_data(read_cycle(port, read_cmd, pic_type),p_or_m,pic_type);

	if (Verbose > 1){
	  printf("Read %s: %04x\n", (p_or_m ? "EEPROM" : "PROG"), readed);
	}

	*data = trunc_data(*data,p_or_m,pic_type);
	if (readed != *data){
	  fprintf(stderr, "Verify error on %s:%04x\nAbort.\n",
			  (p_or_m ? "EEPROM" : "PROG"),cur_addr);
	  exit(1);
	}

	send_sign(port, PIC_CMD_INCR_ADDR);
	data++;
	cur_addr++;
  }
  if (Verbose > 0){
	printf("Done.                                      \n");
  }
}

void pic_read(int * port,
			  int p_or_m,    /* program or data memory */
			  unsigned short * data,  /* buffer */
			  int len,       /* size of buffer */
			  int pic_type)  /* pic device */
{
  int i;
  int read_cmd = PIC_CMD_READ_PROG;
  int cur_addr=0;

  prog_verify_mode(port, ON); /* MCLR=Vpp program/verify mode */

  if (p_or_m == DATA_MEMORY){   /* manip data memory */
	read_cmd = PIC_CMD_READ_DATA;
  } else {
	read_cmd = PIC_CMD_READ_PROG;
  }

  for(i = 0; i < len; i++){
	*data = trunc_data(read_cycle(port, read_cmd, pic_type), p_or_m, pic_type);
	if (Verbose > 1){
	  printf("Address: %04x\n", cur_addr);
	  printf("Read %s: %04x\n", (p_or_m ? "EEPROM" : "PROG"), *data);
	}
	data++;
	cur_addr++;
	send_sign(port, PIC_CMD_INCR_ADDR);
  }
}

void pic_write_fuse(int * port, unsigned short data, int pic_type)
{
  load_cycle(port, 0xaa, PIC_CMD_LOAD_CONFIG, pic_type); /* 0x2000 */
                           /* dummy without BEGIN_PROG */
  send_sign(port, PIC_CMD_INCR_ADDR);  /* 0x2001 */
  send_sign(port, PIC_CMD_INCR_ADDR);  /* 0x2002 */
  send_sign(port, PIC_CMD_INCR_ADDR);  /* 0x2003 */
  send_sign(port, PIC_CMD_INCR_ADDR);  /* 0x2004 */
  send_sign(port, PIC_CMD_INCR_ADDR);  /* 0x2005 */
  send_sign(port, PIC_CMD_INCR_ADDR);  /* 0x2006 */
  send_sign(port, PIC_CMD_INCR_ADDR);  /* 0x2007 */
  data = trunc_data(data, PROGRAM_MEMORY, pic_type);
  if (Verbose > 1){
	printf("Write Fuse: %04x\n", data);
  }
  load_cycle(port, data, PIC_CMD_LOAD_PROG, pic_type);
  send_sign(port, PIC_CMD_BEGIN_PROG);
  pic_wait(10*1000);
  if (trunc_data(read_cycle(port, PIC_CMD_READ_PROG, pic_type),
				 PROGRAM_MEMORY, pic_type) != data){
	fprintf(stderr, "Verify error on Fuse\nAbort.\n");
	exit(1);
  }
}

void pic_chip_erase(int * port, int pic_type)
{
  if (Verbose > 1){
	printf("Begin Chip Erase.\n");
  }
  prog_verify_mode(port, ON);           /* MCLR=Vpp prog/veryfy mode */

  /* BULK ERASE PROGRAM MEMORY */
  load_cycle(port, 0xffff, PIC_CMD_LOAD_PROG, pic_type);
  send_sign(port, PIC_CMD_BULK_ERASE_PROG);
  send_sign(port, PIC_CMD_BEGIN_PROG);
  pic_wait(10*1000);

  /* BULK ERASE PROGRAM DATA */
  load_cycle(port, 0xffff, PIC_CMD_LOAD_DATA, pic_type);
  send_sign(port, PIC_CMD_BULK_ERASE_DATA);
  send_sign(port, PIC_CMD_BEGIN_PROG);
  pic_wait(10*1000);

  prog_verify_mode(port, OFF);
  if (Verbose > 0){
	printf("Chip Erase Done.\n");
  }
}


